/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    MemoryX::ArmarXObjects::GraphImportExport
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <VirtualRobot/VirtualRobot.h>

#include "Visu.h"

#include <RobotAPI/components/ArViz/Client/Client.h>

#include <SimoxUtility/color/Color.h>
#include <SimoxUtility/math/pose.h>


namespace armarx::nav::graph
{

    viz::Pose VertexVisu::Pose::draw(Graph::ConstVertex vertex) const
    {
        const arondto::Vertex& aron = vertex.attrib().aron;
        return viz::Pose(aron.locationID.entityName)
                .pose(aron.globalRobotPose)
                .scale(scale);
    }


    viz::Arrow VertexVisu::ForwardArrow::draw(Graph::ConstVertex vertex) const
    {
        const arondto::Vertex& aron = vertex.attrib().aron;
        return viz::Arrow(aron.locationID.entityName + " forward")
                .fromTo(simox::math::position(aron.globalRobotPose),
                        simox::math::transform_position(aron.globalRobotPose, length * Eigen::Vector3f::UnitY()))
                .color(color)
                .width(width);
    }


    void VertexVisu::draw(viz::Layer& layer, Graph::ConstVertex vertex) const
    {
        if (pose.has_value())
        {
            layer.add(pose->draw(vertex));
        }
        if (forwardArrow.has_value())
        {
            layer.add(forwardArrow->draw(vertex));
        }
    }


    viz::Arrow EdgeVisu::Arrow::draw(Graph::ConstEdge edge) const
    {
        const auto& sourceAron = edge.source().attrib().aron;
        const auto& targetAron = edge.target().attrib().aron;
        return viz::Arrow(sourceAron.locationID.entityName + " -> " +
                          targetAron.locationID.entityName)
                  .fromTo(simox::math::position(sourceAron.globalRobotPose),
                          simox::math::position(targetAron.globalRobotPose))
                  .width(width)
                .color(color);
    }


    void EdgeVisu::draw(viz::Layer& layer, Graph::ConstEdge edge) const
    {
        if (arrow.has_value())
        {
            layer.add(arrow->draw(edge));
        }
    }


    void GraphVisu::draw(viz::Layer& layer, const Graph& graph)
    {
        if (vertex.has_value())
        {
            for (Graph::ConstVertex v : graph.vertices())
            {
                vertex->draw(layer, v);
            }
        }
        if (edge.has_value())
        {
            for (Graph::ConstEdge e : graph.edges())
            {
                edge->draw(layer, e);
            }
        }
    }
}
