/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::gui-plugins::LocationGraphEditorWidgetController
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#pragma once

#include <Navigation/gui-plugins/LocationGraphEditor/ui_LocationGraphEditorWidget.h>


#include <MemoryX/components/PriorKnowledge/PriorKnowledge.h>
#include <MemoryX/interface/gui/GraphVisualizerInterface.h>
#include <MemoryX/interface/components/GraphNodePoseResolverInterface.h>


#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/interface/visualization/DebugDrawerInterface.h>
#include <RobotAPI/libraries/armem/client/forward_declarations.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>

#include <ArmarXGui/libraries/ArmarXGuiBase/ArmarXComponentWidgetController.h>
#include <ArmarXGui/libraries/SimpleConfigDialog/SimpleConfigDialog.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/system/ImportExportComponent.h>

#include <QDialog>
#include <QGraphicsScene>
#include <QGraphicsLineItem>
#include <QGraphicsEllipseItem>
#include <QMainWindow>
#include <QMouseEvent>

#include <string>
#include <map>
#include <vector>
#include <tuple>


namespace armarx
{
    class GraphVisualizerGraphicsEllipseItem;
    class GraphVisualizerGraphicsLineItem;

    class LocationGraphEditorWidgetController;

    class MouseEventProcessor;


    /**
    \page Navigation-GuiPlugins-LocationGraphEditor LocationGraphEditor
    \brief The LocationGraphEditor allows visualizing ...

    \image html LocationGraphEditor.png
    The user can

    API Documentation \ref LocationGraphEditorWidgetController

    \see LocationGraphEditorGuiPlugin
    */


    /**
     * \class LocationGraphEditorGuiPlugin
     * \ingroup ArmarXGuiPlugins
     * \brief LocationGraphEditorGuiPlugin brief description
     *
     * Detailed description
     */

    /**
     * \class LocationGraphEditorWidgetController
     * \brief LocationGraphEditorWidgetController brief one line description
     *
     * Detailed description
     */
    class ARMARXCOMPONENT_IMPORT_EXPORT
        LocationGraphEditorWidgetController :
            public armarx::ArmarXComponentWidgetControllerTemplate < LocationGraphEditorWidgetController >
    {
        Q_OBJECT
        friend class MouseEventProcessor;


    public:

        /// The type of node ids. (This type implies the node exists)
        using NodeId = const std::string;

        /// The type of edge ids. (This type implies the edge exists)
        using EdgeId = const std::pair<const std::string, const std::string>;


        static QString GetWidgetName();
        static QIcon GetWidgetIcon();


        explicit LocationGraphEditorWidgetController();
        virtual ~LocationGraphEditorWidgetController() override;


        //// @see ArmarXWidgetController::loadSettings()
        void loadSettings(QSettings* settings) override;
        /// @see ArmarXWidgetController::saveSettings()
        void saveSettings(QSettings* settings) override;


        QPointer<QDialog> getConfigDialog(QWidget* parent = nullptr) override;
        void configured() override;


        void onInitComponent() override;
        void onConnectComponent() override;


        //because the above load/save functions are *obviously* for "Save/Load Gui Config." :/
        virtual void loadAutomaticSettings();
        virtual void saveAutomaticSettings();


        // slice interface implementation
        bool hasEdge(const std::string& node1, const std::string& node2)
        {
            return (edges.find(toEdge(node1, node2)) != edges.end());
        }
        bool hasNode(const std::string& id)
        {
            return (nodes.find(id) != nodes.end());
        }


    public slots:

        // slice interface implementation
        void addEdge(const std::string& node1Id, const std::string& node2Id);
        void addNode(const memoryx::GraphNodeBasePtr& node);

        void highlightEdge(const std::string& node1Id, const std::string& node2Id, bool highlighted = true);
        void highlightNode(const std::string& nodeId, bool highlighted = true);

        void clearEdges();
        void clearGraph();

        void resetHighlight();

        void redraw();
        void refreshGraph();

        void tableWidgetNodesCustomContextMenu(QPoint pos);
        void tableWidgetEdgesCustomContextMenu(QPoint pos);


    private slots:
        /**
         * @brief add kitchen graph (H2T Armar3a robot kitchen)
         */
        void addKitchenGraph();

        void selectedSceneChanged(int i);
        void updateSceneList();

        void drawScene();

        /**
         * @brief Toggles the double clicked node's selection state.
         * @param row Identifies the node.
         */
        void nodeTableDoubleClicked(int row, int);
        /**
         * @brief Toggles the double clicked edge's selection state.
         * @param row Identifies the edge.
         */
        void edgeTableDoubleClicked(int row, int);

        /**
         * @brief Toggles the double clicked node's selection state.
         * @param id Identifies the node.
         */
        void nodeDoubleClicked(NodeId id);

        /**
         * @brief Toggles the double clicked edge's selection state.
         * @param id Identifies the edge.
         */
        void edgeDoubleClicked(EdgeId id);

        /**
         * @brief Rotates the view clockwise.
         *
         * Stepsize set by VIEW_ROTATE_STEP_SIZE_CC in GraphVisualizerGuiPlugin.cpp
         */
        void viewRotatedClock();

        /**
         * @brief Rotates the view counter clockwise.
         *
         * Stepsize set by VIEW_ROTATE_STEP_SIZE_CC in GraphVisualizerGuiPlugin.cpp
         */
        void viewRotatedCounterClock();

        /**
         * @brief Applies the current transforamtion to the view.
         */
        void transformView();

        /**
         * @brief Adjusts the view's zoom and rotation to display most of the graph.
         */
        void adjustView();

        bool addNewEdge(const std::string& from, const std::string& to);
        void addNewEdgeBoth();
        void addNewEdgeStartEnd();
        void addNewEdgeEndStart();

        void addNewGraphNode();
        void editGraphNode();


    private:

        /// Widget Form
        Ui::LocationGraphEditorWidget widget;


        /**
         * @brief The NodeData struct holds data required for the node.
         * The name is stored in the key used in the map nodes.
         */
        struct NodeData
        {
            /**
             * @brief The Entity of the graph segment this struct represents
             */
            memoryx::GraphNodeBasePtr node;

            /**
             * @brief The pose drawn to debugDrawer.
             */
            armarx::FramedPosePtr pose;

            /**
             * @brief The ellipse in the scene.
             */
            QGraphicsEllipseItem* graphicsItem;

            /**
             * @brief The row in the table tableWidgetNodes.
             */
            int tableWidgetNodesIndex;

            /**
             * @brief Whether the node is highlighted.
             */
            bool highlighted;
        };

        /**
         * @brief The EdgeData struct holds data required for the edge.
         * The name is stored in the key used in the map edges.
         */
        struct EdgeData
        {
            /**
             * @brief The line in the scene.
             */
            QGraphicsLineItem* graphicsItem;
            /**
             * @brief The row in the table tableWidgetEdges.
             */
            int tableWidgetEdgesIndex;

            /**
             * @brief Whether the edge is highlighted.
             */
            bool highlighted;
        };


        /**
         * @brief Returns the EdgeId corresponding to two nodes.
         * @param node1 First node id.
         * @param node2 Second node id.
         * @return The EdgeId corresponding to two nodes.
         */
        static EdgeId toEdge(const std::string& node1, const std::string& node2)
        {
            return EdgeId {node1, node2};
        }

        /**
         * @brief Updates an edge.
         * @param The edge to update.
         */
        void updateEdge(const EdgeId& id);

        /**
         * @brief Updates a node.
         * @param The node to update.
         */
        void updateNode(const NodeId& id);

        void setEditFields(const NodeData& node);

        /**
         * @brief The topic name used by debugDrawer.
         */
        std::string debugDrawerTopicName;

        /**
         * @brief Used to draw onto debug layers.
         */
        ::armarx::DebugDrawerInterfacePrx debugDrawer;

        /**
         * @brief The config dialog.
         */
        QPointer<SimpleConfigDialog> dialog;

        float getYawAngle(const armarx::PoseBasePtr& pose) const;

        /**
         * @brief The scene displayed in the widget.
         *
         * For y coordinates -pos->y is used to mirror the scene on the y axis.
         * If pos->y would be used the graph displayed in the scene would not
         * match the graph drawn to the debug layer.
         */
        QPointer<QGraphicsScene> scene;

        /**
         * @brief The nodes.
         */
        std::map<std::string, NodeData> nodes;

        /**
         * @brief The edges.
         */
        std::map<EdgeId, EdgeData> edges;

        /**
         * @brief The view's rotation angle.
         */
        qreal viewAngle;

        /**
         * @brief The layer to draw on.
         */
        std::string debugDrawerLayerName;

        bool editStartNodeNext;

        std::string priorKnowledgeProxyName;
        memoryx::PriorKnowledgeInterfacePrx priorKnowledgePrx;
        memoryx::GraphNodePoseResolverInterfacePrx gnpr;
        memoryx::GraphMemorySegmentBasePrx graphSeg;
        QSettings settings;
        QString lastSelectedSceneName;

        /**
         * selectScene(): private function called when button load is pushed, and calls the function loadScene()
         */
        void selectScene();

        /**
         * @brief loadScene Private function that parses XML file to load a scene
         * @param xmlFile
         */
        void loadScene(const std::string& xmlFile);


        friend class GraphVisualizerGraphicsEllipseItem;
        friend class GraphVisualizerGraphicsLineItem;


    private:

        std::string memoryNameSystemName = "MemoryNameSystem";
        armem::client::MemoryNameSystem memoryNameSystem;

    };


    /**
     * @brief Required to override the double click event. This is required to toggle the select state.
     */
    class GraphVisualizerGraphicsEllipseItem : public QGraphicsEllipseItem
    {
    public:
        using NodeId = LocationGraphEditorWidgetController::NodeId;

        GraphVisualizerGraphicsEllipseItem(LocationGraphEditorWidgetController& visuWidget, NodeId name, qreal x, qreal y, qreal width, qreal height, QGraphicsItem* parent = nullptr):
            QGraphicsEllipseItem {x, y, width, height, parent},
            id {name},
            parentVisuWidget(visuWidget)
        {
        }

        ~GraphVisualizerGraphicsEllipseItem() override
        {
        }
    protected:
        void mouseDoubleClickEvent(QGraphicsSceneMouseEvent*) override
        {
            parentVisuWidget.nodeDoubleClicked(id);
        }
    private:
        /**
         * @brief Required to identify the element.
         */
        const NodeId id;

        /**
         * @brief Required to call nodeDoubleClicked on it. (This class is no QObject so it does not support signals)
         */
        LocationGraphEditorWidgetController& parentVisuWidget;
    };



    /**
     * @brief Required to override the double click event. This is required to toggle the select state.
     */
    class GraphVisualizerGraphicsLineItem : public QGraphicsLineItem
    {
    public:
        using EdgeId = LocationGraphEditorWidgetController::EdgeId;

        GraphVisualizerGraphicsLineItem(LocationGraphEditorWidgetController& visuWidget, EdgeId name, qreal x1, qreal y1, qreal x2, qreal y2, QGraphicsItem* parent = 0):
            QGraphicsLineItem {x1, y1, x2, y2, parent},
            id {name},
            parentVisuWidget(visuWidget)
        {
        }

        ~GraphVisualizerGraphicsLineItem() override
        {
        }
    signals:
    protected:
        void mouseDoubleClickEvent(QGraphicsSceneMouseEvent*) override
        {
            parentVisuWidget.edgeDoubleClicked(id);
        }
    private:
        /**
         * @brief Required to identify the element.
         */
        const EdgeId id;

        /// Required to call edgeDoubleClicked on it. (This class is no QObject so it does not support signals)
        LocationGraphEditorWidgetController& parentVisuWidget;
    };


    class MouseEventProcessor : public QObject
    {
        Q_OBJECT
    public:

        MouseEventProcessor(LocationGraphEditorWidgetController* gvw) : gvw(gvw) {}


    protected:

        LocationGraphEditorWidgetController* gvw;
        bool eventFilter(QObject* obj, QEvent* event) override;

    };

}
